#include "MyRunAction.hh"

#ifdef G4MULTITHREADED
#include "G4MTRunManager.hh"
#else
#include "G4RunManager.hh"
#endif
#include "G4Timer.hh"
#include "G4SystemOfUnits.hh"
#include "globals.hh"

#include "MyRun.hh"
#include "MyPrimaryGeneratorAction.hh"

#include "G4ProductionCutsTable.hh"
#include "G4Region.hh"
#include "G4RegionStore.hh"

G4AnalysisManager* MyRunAction::fMasterAnalysisManager = nullptr;

MyRunAction::MyRunAction(bool isGeantino, bool isStepStats, G4String geantinoMapFilename, G4String stepStatsFilename) : G4UserRunAction(), fIsPerformance(false), fIsGeantino(isGeantino), fIsStepStats(isStepStats), fRun(nullptr), fTimer(nullptr), fGeantinoMapsFilename(geantinoMapFilename), fStepStatsFilename(stepStatsFilename){
    
}

MyRunAction::~MyRunAction() {
    if(fIsGeantino || fIsStepStats)
        delete G4AnalysisManager::Instance();
}

G4Run* MyRunAction::GenerateRun() {
    // don't Generate our Run in perfomance mode but return with nullptr:
    //   the RunManager::RunInitialization will create a base G4Run object if this method gives back nullptr
    if (!fIsPerformance) {
        fRun = new MyRun();
        return fRun;
    }
    return nullptr;
}


void MyRunAction::BeginOfRunAction(const G4Run* /*aRun*/){
    
    if(fIsGeantino)
    {
        // Create analysis manager
        G4AnalysisManager* analysisManager = G4AnalysisManager::Instance();
        if (isMaster) {
            fMasterAnalysisManager = analysisManager;
            //G4cout<<"MyRunAction::BeginOfRunAction, created MASTER istance of the G4AnalysisManager: "<<fMasterAnalysisManager<<G4endl;
            
        }
        //else
        //    G4cout<<"MyRunAction::BeginOfRunAction, created WORKER istance of the G4AnalysisManager: "<<analysisManager<<G4endl;
        
        G4cout << "Using G4AnalysisManager type: " << analysisManager->GetType() << G4endl;
        analysisManager->SetVerboseLevel(1);
        
        // Open output root file
        G4cout<<"\n\nBeginOfRunAction: \n...create a root file using the G4AnalysisManager" << G4endl;
        if (!analysisManager->OpenFile(fGeantinoMapsFilename)){
            G4cout<<"\nBeginOfRunAction ERROR: File cannot be opened!"<<G4endl;
            exit(-1);
        } else
            G4cout<<"\n...output File "<<fGeantinoMapsFilename<<" opened!"<<G4endl;
        
        const char* radName = "RZRadLen";
        if(analysisManager->GetP2Id(radName, false) < 0){
            fRadName_id = analysisManager->CreateP2(radName,radName,1000,-25000.,25000.,2000,0.,15000.);
            //G4cout<<"MyRunAction::BeginOfRunAction: G4AnalysisManager Created RZRadLen 2DProfile with name: "<<radName<< " and  with id: "<<fRadName_id<<G4endl;
            analysisManager->SetP2XAxisTitle(fRadName_id,"Z[mm]");
            analysisManager->SetP2YAxisTitle(fRadName_id,"R[mm]");
            analysisManager->SetP2ZAxisTitle(fRadName_id,"thickstepRL");
            
        }
        const char* intName = "RZIntLen";
        if(analysisManager->GetP2Id(intName, false)< 0)
        {
            fIntName_id = analysisManager->CreateP2(intName,intName,1000,-25000.,25000.,2000,0.,15000.);
            //G4cout<<"MyRunAction::BeginOfRunAction: G4AnalysisManager Created RZIntLen 2DProfile with name: "<<intName<< " and with id: "<<fIntName_id<<G4endl;
            analysisManager->SetP2XAxisTitle(fIntName_id,"Z[mm]");
            analysisManager->SetP2YAxisTitle(fIntName_id,"R[mm]");
            analysisManager->SetP2ZAxisTitle(fIntName_id,"thickstepIL");
        }
        const char* etaRadName = "etaRadLen";
        if(analysisManager->GetP1Id(etaRadName, false)< 0)
        {
            fEtaRad_id = analysisManager->CreateP1(etaRadName,etaRadName,500,-6,6);
            analysisManager->SetP1XAxisTitle(fEtaRad_id,"#eta");
            analysisManager->SetP1YAxisTitle(fEtaRad_id,"Radiation Length %X0");
           
        }
        const char* etaIntName = "etaIntLen";
        if(analysisManager->GetP1Id(etaIntName, false)< 0)
        {
            fEtaInt_id = analysisManager->CreateP1(etaIntName,etaIntName,500,-6,6);
            analysisManager->SetP1XAxisTitle(fEtaInt_id,"#eta");
            analysisManager->SetP1YAxisTitle(fEtaInt_id,"Interaction Length #lambda");
        }
        
    }

    if(fIsStepStats)
    {
        // Create analysis manager
        G4AnalysisManager* analysisManager = G4AnalysisManager::Instance();
        if (isMaster) {
            fMasterAnalysisManager = analysisManager;
            //G4cout<<"MyRunAction::BeginOfRunAction, created MASTER istance of the G4AnalysisManager: "<<fMasterAnalysisManager<<G4endl;
        }
        //else
        //    G4cout<<"MyRunAction::BeginOfRunAction, created WORKER istance of the G4AnalysisManager: "<<analysisManager<<G4endl;
        
        G4cout << "Using G4AnalysisManager type: " << analysisManager->GetType() << G4endl;
        analysisManager->SetVerboseLevel(1);
        
        // Open output root file
        G4cout<<"\n\nBeginOfRunAction: \n...create a root file using the G4AnalysisManager" << G4endl;
        if (!analysisManager->OpenFile(fStepStatsFilename)) {
            G4cout<<"\nBeginOfRunAction ERROR: File cannot be opened!"<<G4endl;
            exit(-1);
        }
        else
            G4cout<<"\n...output File "<<fStepStatsFilename<<" opened!"<<G4endl;
        
        const char* stepHeatMapName = "ZRSteps";
        if(analysisManager->GetH2Id(stepHeatMapName, false) < 0) {
            fStepHeatMap_id = analysisManager->CreateH2(stepHeatMapName,stepHeatMapName,1600,-4000.,4000.,2000,0.,5000.);
            analysisManager->SetH2XAxisTitle(fStepHeatMap_id,"Z [mm]");
            analysisManager->SetH2YAxisTitle(fStepHeatMap_id,"R [mm]");
            analysisManager->SetH2ZAxisTitle(fStepHeatMap_id,"Steps");
        }

        const char* stepKineticName = "ESteps";
        if(analysisManager->GetH1Id(stepKineticName, false) < 0) {
            fStepKinetic_id = analysisManager->CreateH1(stepKineticName,stepKineticName,1000,-9,7);
            analysisManager->SetH1XAxisTitle(fStepKinetic_id, "log( pre-step kinetic energy [MeV] )");
            analysisManager->SetH1YAxisTitle(fStepKinetic_id, "Steps");
        }

        const char* stepLengthName = "LSteps";
        if(analysisManager->GetH1Id(stepLengthName, false) < 0) {
            fStepLength_id = analysisManager->CreateH1(stepLengthName,stepLengthName,1000,-12,4);
            analysisManager->SetH1XAxisTitle(fStepLength_id, "log( step length [mm] )");
            analysisManager->SetH1YAxisTitle(fStepLength_id, "Steps");
        }

        const char* stepLengthInName = "LStepsIn";
        if(analysisManager->GetH1Id(stepLengthInName, false) < 0) {
            fStepLengthIn_id = analysisManager->CreateH1(stepLengthInName,stepLengthInName,1000,-12,4);
            analysisManager->SetH1XAxisTitle(fStepLengthIn_id, "log( step length [mm] )");
            analysisManager->SetH1YAxisTitle(fStepLengthIn_id, "Steps");
        }

        const char* stepLengthOutName = "LStepsOut";
        if(analysisManager->GetH1Id(stepLengthOutName, false) < 0) {
            fStepLengthOut_id = analysisManager->CreateH1(stepLengthOutName,stepLengthOutName,1000,-12,4);
            analysisManager->SetH1XAxisTitle(fStepLengthOut_id, "log( step length [mm] )");
            analysisManager->SetH1YAxisTitle(fStepLengthOut_id, "Steps");
        }

        const char* trackInitKineticName = "InitETracks";
        if(analysisManager->GetH1Id(trackInitKineticName, false) < 0) {
            fInitTrackKinetic_id = analysisManager->CreateH1(trackInitKineticName,trackInitKineticName,1000,-9,7);
            analysisManager->SetH1XAxisTitle(fInitTrackKinetic_id, "log( initial kinetic energy [MeV] )");
            analysisManager->SetH1YAxisTitle(fInitTrackKinetic_id, "Tracks");
        }

        const char* numOfStepsName = "NumOfSteps";
        if(analysisManager->GetH1Id(numOfStepsName, false) < 0) {
            fNumOfSteps_id = analysisManager->CreateH1(numOfStepsName,numOfStepsName,10000,0.5,10000.5);
            analysisManager->SetH1XAxisTitle(fNumOfSteps_id, "Number of steps");
            analysisManager->SetH1YAxisTitle(fNumOfSteps_id, "Tracks");
        }

        const char* numOfStepsPerInitEName = "NumOfStepsPerInitE";
        if(analysisManager->GetH1Id(numOfStepsPerInitEName, false) < 0) {
            fNumOfStepsPerInitE_id = analysisManager->CreateH1(numOfStepsPerInitEName,numOfStepsPerInitEName,1000,-9,7);
            analysisManager->SetH1XAxisTitle(fNumOfStepsPerInitE_id, "log( step-weighted initial kinetic energy [MeV] )");
            analysisManager->SetH1YAxisTitle(fNumOfStepsPerInitE_id, "Tracks");
        }

        const char* numOfBoundXingsName = "NumOfBoundXings";
        if(analysisManager->GetH1Id(numOfBoundXingsName, false) < 0) {
            fNumOfBoundXings_id = analysisManager->CreateH1(numOfBoundXingsName,numOfBoundXingsName,1000,0.5,1000.5);
            analysisManager->SetH1XAxisTitle(fNumOfBoundXings_id, "Number of boundary crossings");
            analysisManager->SetH1YAxisTitle(fNumOfBoundXings_id, "Tracks");
        }
    }
    
    if (isMaster) {
        
        //G4cout<<"\nBeginOfRunAction isMaster, and fMasterAnalysisManager: "<<fMasterAnalysisManager<<G4endl;
        
        std::vector<G4Region*>* regionVect =  G4RegionStore::GetInstance();
        int numRegions = regionVect->size();
        int sumNumMC = 0;
        G4ProductionCutsTable* pTable = G4ProductionCutsTable::GetProductionCutsTable();
        size_t numMC = pTable->GetTableSize();
        std::vector<int> mcRVect(23,0);
        for (size_t imc=0; imc<numMC; ++imc) {
            const G4MaterialCutsCouple* mc =pTable->GetMaterialCutsCouple(imc);
            if (!mc->IsUsed()) continue;
            G4Material* mat = const_cast<G4Material*>(mc->GetMaterial());
            int k=0;
            for (; k<numRegions && (!(*regionVect)[k]->FindCouple(mat)); ++k){}
            if (k<numRegions) {
                ++mcRVect[k];
            }
        }
        for (int ir=0; ir<numRegions; ++ir) {
            G4cout<< " ir = " << ir << "  region Name = " << (*regionVect)[ir]->GetName() <<" #mc = " << mcRVect[ir] << G4endl;
            sumNumMC += mcRVect[ir];
        }
        G4cout<< " === Total number of MC = " << sumNumMC << " vs " << numMC << " #regions = " << numRegions << G4endl;
        
        
#ifdef G4MULTITHREADED
        G4int nThreads = G4MTRunManager::GetMasterRunManager()->GetNumberOfThreads();
        G4cout << "\n  =======================================================================================\n"
        << "   Run started in MT mode with " << nThreads << "  threads    \n"
        << "  =======================================================================================  \n"
        << G4endl;
#else
        G4cout << "\n  =======================================================================================\n"
        << "   Run started in sequential mode (i.e. with 1 thread)        \n"
        << "  =======================================================================================  \n"
        << G4endl;
#endif
        fTimer = new G4Timer();
        fTimer->Start();
    }
    //else G4cout<<"BeginOfRunAction isWorker, and fMasterAnalysisManager: "<<fMasterAnalysisManager<<G4endl;
}


void MyRunAction::EndOfRunAction(const G4Run*) {
    
    if(fIsGeantino || fIsStepStats){
        
        auto analysisManager= G4AnalysisManager::Instance();
        //Finalize analysisManager and Write out file
        if (analysisManager->IsOpenFile()){
            G4cout<<"\n\n EndOfRunAction, writing output file: "<<analysisManager->GetFileName() << G4endl;
            analysisManager->Write();
            G4cout<<"... closing output file: "<<analysisManager->GetFileName() << G4endl;
            analysisManager->CloseFile();
            G4cout<<"Output File successfully saved and closed! " << G4endl;
        }
        
    }
    if (isMaster) {
        fTimer->Stop();
        // get number of events: even in case of perfomance mode when MyRun-s are not generated in GenerateRun()
#ifdef G4MULTITHREADED
        const G4Run* run = G4MTRunManager::GetMasterRunManager()->GetCurrentRun();
#else
        const G4Run* run = G4RunManager::GetRunManager()->GetCurrentRun();
#endif
        G4int nEvt   = run->GetNumberOfEvent();
        G4cout << "\n\n" << G4endl;
        G4cout << "  =======================================================================================  " << G4endl;
        G4cout << "   Run terminated:                                                                         " << G4endl;
        G4cout << "     Number of events transported    = " << nEvt                                             << G4endl;
        G4cout << "     Time:  "  << *fTimer                                                                   << G4endl;
        G4cout << "  =======================================================================================  "<< G4endl;
        delete fTimer;
        // print primary gun properties (not available at the begining of the run)
        MyPrimaryGeneratorAction::Print();
        if (!fIsPerformance) { // otherwise we do not even create any MyRun objects so fRun is nullptr
            fRun->EndOfRun();
        }
    }
}
