
#include "MySteppingAction.hh"

#include "MyEventAction.hh"
#include "MyRunAction.hh"
#include "MyTrackInformation.hh"
#include "G4Step.hh"
#include "G4ParticleDefinition.hh"
#include "G4FieldManager.hh"
#include "G4Field.hh"
#include <cmath>


MySteppingAction::MySteppingAction(MyEventAction* evtact, MyRunAction* runact) : G4UserSteppingAction(), fEventAction(evtact), fRunAction(runact) {}


MySteppingAction::~MySteppingAction() {}


void MySteppingAction::UserSteppingAction(const G4Step* theStep) {
  G4double edep      = theStep->GetTotalEnergyDeposit();
  G4double stepl     = theStep->GetStepLength();
  G4Track* theTrack  = theStep->GetTrack();
  G4bool   isCharged = (theTrack->GetDefinition()->GetPDGCharge()!=0.);
  G4int primTrackID  = static_cast<MyTrackInformation*>(theTrack->GetUserInformation())->GetPrimaryTrackID();
  fEventAction->AddData(edep, stepl, isCharged, primTrackID-1);

  /* read and print field

  const G4int stepNumber = theTrack->GetCurrentStepNumber();
  std::cout<<"CurrentStepNumber is: "<<stepNumber<<std::endl;

  auto preStepPoint = theStep->GetPreStepPoint();
  auto position = preStepPoint->GetPosition();
  const double point[3] = {position.getX(), position.getY(), position.getZ()};

  auto volume = preStepPoint->GetPhysicalVolume()->GetLogicalVolume();
  auto fieldManager = volume->GetFieldManager();
  if (fieldManager) {
  	const G4Field* field = fieldManager->GetDetectorField();
  	if (field) {
  		std::cout<<"Field DOES exist!"<<std::endl;
  		double fieldValue[3];
  		field->GetFieldValue(point, fieldValue);
  		std::cout<<"Field value is: "<<fieldValue[0]<<", "<<fieldValue[1]<<", "<<fieldValue[2]<<std::endl;
  	}
  	else 
  		std::cout<<"Field does NOT exist!"<<std::endl;
  }
  else 
  	std::cout<<"Field manager does NOT exist!"<<std::endl;
<<<<<<< HEAD
  
  */

  // Collect step statistics
  G4bool doStepStats = fRunAction->getIsStepStats();
  if (doStepStats) {
    
    // pre-step point
    const G4StepPoint* PreStepPoint = theStep->GetPreStepPoint();
    // post-step point
    const G4StepPoint* PostStepPoint = theStep->GetPostStepPoint();
    // pre-step position
    const G4ThreeVector position = PreStepPoint->GetPosition();
    const G4double rXY = position.rho();
    const G4double Z = position.z();

    const G4double r2XY = position.x()*position.x() + position.y()*position.y();
    const G4double dOffset = 100.0; // 'zero field area' dOffset mm away of the Calo boudaries
    const G4double r2Min = (1500 + dOffset)*(1500 + dOffset);
    const G4double r2Max = (3700 - dOffset)*(3700 - dOffset);
    const bool inSide = (r2XY > r2Min) && (r2XY < r2Max) && (abs(Z) < 1000 - dOffset);

    // pre-step kinetic energy
    const G4double stepKinetic = PreStepPoint->GetKineticEnergy();
    // first step (after initial step)
    if (theTrack->GetCurrentStepNumber() == 1) {
      // save track ID in the first step
      fTrackID = theTrack->GetTrackID();
      // reset boundary crossings counter
      fBoundXings = 0;
      // initial kinetic energy
      fInitStepKinetic = stepKinetic;
    }

    // check boundary crossings
    bool boundaryXing = (not (theStep->GetTrack()->GetVolume() == theStep->GetTrack()->GetNextVolume()));
    if (boundaryXing)
      fBoundXings++;

    // last step
    if (theTrack->GetTrackStatus() == 2) {
      // assert to check if we have the correct track
      if (not (theTrack->GetTrackID() == fTrackID)) {
        G4cout << "Track ID changed between the assumed first step and the last." << G4endl;
        throw std::exception();
      }
      // number of steps
      fNSteps = theTrack->GetCurrentStepNumber() + 1;
    }

    // Fill histograms
    auto analysisManager = G4AnalysisManager::Instance();
    {
      // Protect concurrent histo filling
      static std::mutex mutex_instance;
      std::lock_guard<std::mutex> lock(mutex_instance);

      // steps in Z/R
      analysisManager->FillH2(fRunAction->fStepHeatMap_id, Z, rXY);
      // step kinetic energy
      analysisManager->FillH1(fRunAction->fStepKinetic_id, log10(stepKinetic));
      // step length
      analysisManager->FillH1(fRunAction->fStepLength_id, log10(stepl));
      if (inSide)
        analysisManager->FillH1(fRunAction->fStepLengthIn_id, log10(stepl));
      else
        analysisManager->FillH1(fRunAction->fStepLengthOut_id, log10(stepl));

      // first step (after initial step)
      if (theTrack->GetCurrentStepNumber() == 1)
        analysisManager->FillH1(fRunAction->fInitTrackKinetic_id, log10(fInitStepKinetic));

      // last step
      if (theTrack->GetTrackStatus() == 2) {
        // number of steps
        analysisManager->FillH1(fRunAction->fNumOfSteps_id, fNSteps);
        // number of steps vs initial energy
        analysisManager->FillH1(fRunAction->fNumOfStepsPerInitE_id, log10(fInitStepKinetic), fNSteps);
        // number of boundary crossings
        analysisManager->FillH1(fRunAction->fNumOfBoundXings_id, fBoundXings);
      }

    }
  }
}