/*
  Copyright (C) 2002-2020 CERN for the benefit of the ATLAS collaboration
*/

#include "G4ThreeVector.hh"
#include "G4Field.hh"

#include "SwitchingFieldManager.hh"
#include "G4ChordFinder.hh"

#include "G4Track.hh"
#include "G4MuonPlus.hh"
#include "G4MuonMinus.hh"

SwitchingFieldManager::SwitchingFieldManager( G4Field* detectorField )
   :  G4FieldManager( detectorField ),
      m_savedField( detectorField )
{}

SwitchingFieldManager::~SwitchingFieldManager()
{
    std::cout<<"Destructor SwitchingFieldManager. \nStatistics:"<<std::endl;
    std::cout<<"N. of times turned field OFF: "<<counterOFF<<std::endl;
    std::cout<<"N. of times turned field ON: "<<counterON<<std::endl;
    std::cout<<"N. of times do nothing INSIDE: "<<counterNothingInside<<std::endl;
    std::cout<<"N. of times do nothing OUTSIDE: "<<counterNothingOutside<<std::endl;
    std::cout<<"N. of WTF: "<<counterWTF<<std::endl;
    std::cout<<"N. of times field exists: "<<counterFieldExist<<std::endl;
    std::cout<<"N. of times field does not exist: "<<counterFieldNotExist<<std::endl;
    std::cout<<"============== "<<std::endl;
}

//  Inside Calo area switch off the B-field.
//
void SwitchingFieldManager::ConfigureForTrack(const G4Track * track)
{
   assert(  GetDetectorField() == m_savedField
         || GetDetectorField() == nullptr );
   // Must ensure that SetDetectorField() has NOT been called to establish
   //   a different field by an inside entity
   //   ... it would invalidate our assumptions !

   const bool isMuonTrack = (track->GetDefinition()==G4MuonPlus::Definition() ||
                             track->GetDefinition()==G4MuonMinus::Definition() );

   const G4ThreeVector position = track->GetPosition();

   // const G4double  rXY = position.rho(); // or .perp() sqrt(x*x + y*y) // according to JA this is slow
   const G4double r2XY = position.x()*position.x() + position.y()*position.y(); 
   const G4double Z = position.z();
   const G4double dOffset = 100.0; // 'zero field area' dOffset mm away of the Calo boudaries

   const G4double r2Min = (m_radiusXYmin + dOffset)*(m_radiusXYmin + dOffset);
   const G4double r2Max = (m_radiusXYmax - dOffset)*(m_radiusXYmax - dOffset);

   // const bool outSide = rXY > m_radiusXYmax; outSide -> inSide
   const bool inSide = (r2XY > r2Min) && (r2XY < r2Max) && (abs(Z) < m_Zmax - dOffset);

   if ( inSide && DoesFieldExist() && !isMuonTrack) {
    ChangeDetectorField( nullptr );
   }
   else if ( !inSide && !DoesFieldExist() ) {
    ChangeDetectorField( m_savedField );
   }

   /* sanity checks

   if(DoesFieldExist()) counterFieldExist++;
   else counterFieldNotExist++;

   if ( !inSide && DoesFieldExist() ) {
    // std::cout<<"Field is ON. I do NOTHING!"<<std::endl;
    counterNothingOutside++;
   }
   else if ( !inSide && !DoesFieldExist() ) {
    // std::cout<<"Field has been turned ON!"<<std::endl;
    ChangeDetectorField( m_savedField );
    counterON++;
   }
   else if ( inSide && DoesFieldExist() ) {
    // std::cout<<"Field has been turned OFF!"<<std::endl;
    ChangeDetectorField( nullptr );
    counterOFF++;
   }
   else if ( inSide && !DoesFieldExist() ) {
    // std::cout<<"Field is OFF. I do NOTHING!"<<std::endl;
    counterNothingInside++;
  }
  else counterWTF++;
  */
}


G4FieldManager* SwitchingFieldManager::Clone() const
{
   G4Field* cloneField = m_savedField->Clone();
   auto clone= new SwitchingFieldManager( cloneField );

   clone->SetMinimumEpsilonStep( this->GetMinimumEpsilonStep() );
   clone->SetMaximumEpsilonStep( this->GetMaximumEpsilonStep() );
   clone->SetDeltaOneStep(       this->GetDeltaOneStep() );
   clone->SetDeltaIntersection(  this->GetDeltaIntersection()  );

   clone->GetChordFinder()->SetDeltaChord( this->GetChordFinder()->GetDeltaChord() );
   // This lives in our chord finder ...
   
   return clone;
}

